
from BeautifulSoup import BeautifulSoup
from handle_html import Node
import re

class CSRFTestor:
    """ Class to check if the forms are protected against CSRF
    @author rmillet
    """

    tree = None
    url = None
    cookie = None
    vulnerable = []
    notVulnerable = []
    bs = []

    def __init__(self, url, cookie, final_tree=None, html_strings=None):
        self.tree = final_tree
        self.url = url
        self.cookie = cookie

        if self.tree != None:
            for i in range(len(self.tree.child)):
                self.bs.append(BeautifulSoup(self.treeToString(final_tree.child[i])))
        elif html_strings != None:
            for i in range(len(html_strings)):
                self.bs.append(BeautifulSoup(html_strings[i]))

        self.reHidden = re.compile("^hidden$", re.I)
        self.rePassword = re.compile("^password$", re.I)
        self.reSubmit = re.compile("^submit$", re.I)
        self.reNotempty = re.compile("^.+$")

        self.reHexa = re.compile("^[0-9a-f]{8,}|[0-9A-F]{8,}$")
        self.reBase64 = re.compile("^[A-Za-z0-9+/=]{8,}$")
        self.reUuid = re.compile("^[A-Fa-f0-9]{8}-[A-Fa-f0-9]{4}-[A-Fa-f0-9]{4}-[A-Fa-f0-9]{4}-[A-Fa-f0-9]{12}$")

    def getVulnerableTreeForms(self):
        """ All non-reported form may be false-negative
        Be sure to invoke *Detection methods before this method
        """
        result = {}
        result['Vulnerable'] = Node("root")
        result['NotVulnerable'] = Node("root")
        result['Unknown'] = Node("root")
        for i in range(len(self.bs)):
            if i in self.vulnerable:
                result['Vulnerable'].add_child(node=self.tree.child[i])
            elif i in self.notVulnerable:
                result['NotVulnerable'].add_child(node=self.tree.child[i])
            else:
                result['Unknown'].add_child(node=self.tree.child[i])
        return result

    def getVulnerableForms(self):
        """ All non-reported form may be false-negative
        Be sure to invoke *Detection methods before this method
        """
        result = {}
        result['Vulnerable'] = []
        result['NotVulnerable'] = []
        result['Unknown'] = []
        for i in range(len(self.bs)):
            self.bs[i].decompose()
            if i in self.vulnerable:
                result['Vulnerable'].append(self.bs[i])
            elif i in self.notVulnerable:
                result['NotVulnerable'].append(self.bs[i])
            else:
                result['Unknown'].append(self.bs[i])
        return result


    def simpleDetection(self):
        """ If this has no hidden (with name and value) and no password field,
        declare the form vulnerable
        WARNING: this is a False Positive when there is a captcha
        """
        for i in range(len(self.bs)):
            if i in self.vulnerable or i in self.notVulnerable:
                continue
            # No submit, here for decoration
            if (len(self.bs[i].findAll("input", attrs={ 'type' : self.reSubmit })) == 0 and len(self.bs[i].findAll("button", attrs={ 'type' : self.reSubmit })) == 0):
                self.notVulnerable.append(i)
            elif (len(self.bs[i].findAll("input", attrs={ 'type' : self.reHidden, 'name' : self.reNotempty , 'value' : self.reNotempty })) == 0
                and len(self.bs[i].findAll("input", type=self.rePassword)) == 0):
                self.vulnerable.append(i)
            

    def tokenDetection(self):
        """ Try to detect what could be a token """
        for i in range(len(self.bs)):
            if i in self.vulnerable or i in self.notVulnerable:
                continue
            
            hasToken = False
            hidden = self.bs[i].findAll("input", attrs={ 'type' : self.reHidden, 'name' : self.reNotempty, 'value' : self.reNotempty })

            for j in range(len(hidden)):
                hidden[j]['isToken?'] = self.isToken(hidden[j]['name'], hidden[j]['value'])
                if hidden[j]['isToken?'] == True:
                    hasToken = True
                    
                #print hidden[j]

            if hasToken == True:
                self.notVulnerable.append(i)

    def isToken(self, name, value):
        if name in [ "__FORM_TOKEN", "__REQUESTDIGEST", "form_token" ] and len(value) >= 8:
            return True

        # Hexa (md5, sha1, ...)
        if self.reHexa.match(value) != None:
            return True

        # Base64
        if self.reBase64.match(value) != None and len(value) % 4 == 0:
            return True

        # UUID
        if self.reUuid.match(value) != None:
            return True

        return False

    def treeToString(self, tree, printAll=False):
        """ Tranform the tree into a string which contains only
        form html (form, input, ...)
        """
        s = ""
        tag = tree.value.lower()[1:].split(" ", 2)[0]

        if printAll == True or tag in ("form", "input", "select", "button"):
            s += tree.value
        elif tag in ("option", "textarea"):
            s += tree.value
            printAll = True

        for i in range(len(tree.child)):
            s += self.treeToString(tree.child[i], printAll)

        if tag in ("form", "select", "option", "textarea"):
            s += "</" + tag + ">"

        return s


if  __name__ == "__main__":
    exit(42)
